@@ -32,7 +32,8 @@ for dir in "${reposdir}/"*/; do
 
 	mkdir -p "${destdir}/${d}"
 	cd "${destdir}/${d}" || continue
-	stagit-gopher -b "${stagitdir}/${d}" -c ".cache" "${reposdir}/${r}"
+	stagit-gopher -b "${stagitdir}/${d}" -c ".cache" \
+		-u "gopher://codemadness.org/1/git/$d/" "${reposdir}/${r}"
 
 	# symlinks
 	ln -sf log.gph index.gph
