@@ -10,6 +10,9 @@
 
 #include <git2.h>
 
+#define PAD_TRUNCATE_SYMBOL    "\xe2\x80\xa6" /* symbol: "ellipsis" */
+#define UTF_INVALID_SYMBOL     "\xef\xbf\xbd" /* symbol: "replacement" */
+
 static git_repository *repo;
 
 static const char *relpath = "";
@@ -17,36 +20,61 @@ static const char *relpath = "";
 static char description[255] = "Repositories";
 static char *name = "";
 
-/* format `len' columns of characters. If string is shorter pad the rest
+/* Format `len' columns of characters. If string is shorter pad the rest
  * with characters `pad`. */
 int
 utf8pad(char *buf, size_t bufsiz, const char *s, size_t len, int pad)
 {
-	wchar_t w;
+	wchar_t wc;
 	size_t col = 0, i, slen, siz = 0;
-	int rl, wc;
+	int inc, rl, w;
 
-	if (!len)
+	if (!bufsiz)
 		return -1;
+	if (!len) {
+		buf[0] = '\0';
+		return 0;
+	}
 
 	slen = strlen(s);
-	for (i = 0; i < slen && col < len + 1; i += rl) {
-		if ((rl = mbtowc(&w, &s[i], slen - i < 4 ? slen - i : 4)) <= 0)
-			break;
-		if ((wc = wcwidth(w)) == -1)
-			wc = 1;
-		col += wc;
-		if (col >= len && s[i + rl]) {
+	for (i = 0; i < slen; i += inc) {
+		inc = 1; /* next byte */
+		if ((unsigned char)s[i] < 32)
+			continue;
+
+		rl = mbtowc(&wc, &s[i], slen - i < 4 ? slen - i : 4);
+		inc = rl;
+		if (rl < 0) {
+			mbtowc(NULL, NULL, 0); /* reset state */
+			inc = 1; /* invalid, seek next byte */
+			w = 1; /* replacement char is one width */
+		} else if ((w = wcwidth(wc)) == -1) {
+			continue;
+		}
+
+		if (col + w > len || (col + w == len && s[i + inc])) {
 			if (siz + 4 >= bufsiz)
 				return -1;
-			memcpy(&buf[siz], "\xe2\x80\xa6", 4);
-			return 0;
+			memcpy(&buf[siz], PAD_TRUNCATE_SYMBOL, sizeof(PAD_TRUNCATE_SYMBOL) - 1);
+			siz += sizeof(PAD_TRUNCATE_SYMBOL) - 1;
+			buf[siz] = '\0';
+			col++;
+			break;
+		} else if (rl < 0) {
+			if (siz + 4 >= bufsiz)
+				return -1;
+			memcpy(&buf[siz], UTF_INVALID_SYMBOL, sizeof(UTF_INVALID_SYMBOL) - 1);
+			siz += sizeof(UTF_INVALID_SYMBOL) - 1;
+			buf[siz] = '\0';
+			col++;
+			continue;
 		}
-		if (siz + rl + 1 >= bufsiz)
+		if (siz + inc + 1 >= bufsiz)
 			return -1;
-		memcpy(&buf[siz], &s[i], rl);
-		siz += rl;
+		memcpy(&buf[siz], &s[i], inc);
+		siz += inc;
 		buf[siz] = '\0';
+		col += w;
 	}
 
 	len -= col;
@@ -75,7 +103,7 @@ gphtext(FILE *fp, const char *s, size_t len)
 			fputs("        ", fp);
 			break;
 		default:
-			fputc(*s, fp);
+			putc(*s, fp);
 			break;
 		}
 	}
@@ -99,7 +127,7 @@ gphlink(FILE *fp, const char *s, size_t len)
 			fputs("\\|", fp);
 			break;
 		default:
-			fputc(*s, fp);
+			putc(*s, fp);
 			break;
 		}
 	}
@@ -214,10 +242,6 @@ main(int argc, char *argv[])
 	git_libgit2_init();
 
 #ifdef __OpenBSD__
-	for (i = 1; i < argc; i++)
-		if (unveil(argv[i], "r") == -1)
-			err(1, "unveil: %s", argv[i]);
-
 	if (pledge("stdio rpath", NULL) == -1)
 		err(1, "pledge");
 #endif
