@@ -18,6 +18,10 @@
 
 #include "compat.h"
 
+#define LEN(s)    (sizeof(s)/sizeof(*s))
+#define PAD_TRUNCATE_SYMBOL    "\xe2\x80\xa6" /* symbol: "ellipsis" */
+#define UTF_INVALID_SYMBOL     "\xef\xbf\xbd" /* symbol: "replacement" */
+
 struct deltainfo {
 	git_patch *patch;
 
@@ -50,8 +54,15 @@ struct commitinfo {
 	size_t ndeltas;
 };
 
+/* reference and associated data for sorting */
+struct referenceinfo {
+	struct git_reference *ref;
+	struct commitinfo *ci;
+};
+
 static git_repository *repo;
 
+static const char *baseurl = ""; /* base URL to make absolute RSS/Atom URI */
 static const char *relpath = "";
 static const char *repodir;
 
@@ -72,40 +83,61 @@ static char lastoidstr[GIT_OID_HEXSZ + 2]; /* id + newline + NUL byte */
 static FILE *rcachefp, *wcachefp;
 static const char *cachefile;
 
-/* format `len' columns of characters. If string is shorter pad the rest
+/* Format `len' columns of characters. If string is shorter pad the rest
  * with characters `pad`. */
 int
 utf8pad(char *buf, size_t bufsiz, const char *s, size_t len, int pad)
 {
 	wchar_t wc;
 	size_t col = 0, i, slen, siz = 0;
-	int rl, w;
+	int inc, rl, w;
 
-	if (!len)
+	if (!bufsiz)
 		return -1;
+	if (!len) {
+		buf[0] = '\0';
+		return 0;
+	}
 
 	slen = strlen(s);
-	for (i = 0; i < slen; i += rl) {
-		if ((rl = mbtowc(&wc, &s[i], slen - i < 4 ? slen - i : 4)) <= 0)
-			break;
-		if ((w = wcwidth(wc)) == -1)
+	for (i = 0; i < slen; i += inc) {
+		inc = 1; /* next byte */
+		if ((unsigned char)s[i] < 32)
 			continue;
-		if (col + w > len || (col + w == len && s[i + rl])) {
+
+		rl = mbtowc(&wc, &s[i], slen - i < 4 ? slen - i : 4);
+		inc = rl;
+		if (rl < 0) {
+			mbtowc(NULL, NULL, 0); /* reset state */
+			inc = 1; /* invalid, seek next byte */
+			w = 1; /* replacement char is one width */
+		} else if ((w = wcwidth(wc)) == -1) {
+			continue;
+		}
+
+		if (col + w > len || (col + w == len && s[i + inc])) {
 			if (siz + 4 >= bufsiz)
 				return -1;
-			memcpy(&buf[siz], "\xe2\x80\xa6", 3);
-			siz += 3;
-			if (col + w == len && w > 1)
-				buf[siz++] = pad;
+			memcpy(&buf[siz], PAD_TRUNCATE_SYMBOL, sizeof(PAD_TRUNCATE_SYMBOL) - 1);
+			siz += sizeof(PAD_TRUNCATE_SYMBOL) - 1;
 			buf[siz] = '\0';
-			return 0;
+			col++;
+			break;
+		} else if (rl < 0) {
+			if (siz + 4 >= bufsiz)
+				return -1;
+			memcpy(&buf[siz], UTF_INVALID_SYMBOL, sizeof(UTF_INVALID_SYMBOL) - 1);
+			siz += sizeof(UTF_INVALID_SYMBOL) - 1;
+			buf[siz] = '\0';
+			col++;
+			continue;
 		}
-		if (siz + rl + 1 >= bufsiz)
+		if (siz + inc + 1 >= bufsiz)
 			return -1;
-		memcpy(&buf[siz], &s[i], rl);
-		col += w;
-		siz += rl;
+		memcpy(&buf[siz], &s[i], inc);
+		siz += inc;
 		buf[siz] = '\0';
+		col += w;
 	}
 
 	len -= col;
@@ -288,13 +320,110 @@ err:
 	return NULL;
 }
 
+int
+refs_cmp(const void *v1, const void *v2)
+{
+	const struct referenceinfo *r1 = v1, *r2 = v2;
+	time_t t1, t2;
+	int r;
+
+	if ((r = git_reference_is_tag(r1->ref) - git_reference_is_tag(r2->ref)))
+		return r;
+
+	t1 = r1->ci->author ? r1->ci->author->when.time : 0;
+	t2 = r2->ci->author ? r2->ci->author->when.time : 0;
+	if ((r = t1 > t2 ? -1 : (t1 == t2 ? 0 : 1)))
+		return r;
+
+	return strcmp(git_reference_shorthand(r1->ref),
+	              git_reference_shorthand(r2->ref));
+}
+
+int
+getrefs(struct referenceinfo **pris, size_t *prefcount)
+{
+	struct referenceinfo *ris = NULL;
+	struct commitinfo *ci = NULL;
+	git_reference_iterator *it = NULL;
+	const git_oid *id = NULL;
+	git_object *obj = NULL;
+	git_reference *dref = NULL, *r, *ref = NULL;
+	size_t i, refcount;
+
+	*pris = NULL;
+	*prefcount = 0;
+
+	if (git_reference_iterator_new(&it, repo))
+		return -1;
+
+	for (refcount = 0; !git_reference_next(&ref, it); ) {
+		if (!git_reference_is_branch(ref) && !git_reference_is_tag(ref)) {
+			git_reference_free(ref);
+			ref = NULL;
+			continue;
+		}
+
+		switch (git_reference_type(ref)) {
+		case GIT_REF_SYMBOLIC:
+			if (git_reference_resolve(&dref, ref))
+				goto err;
+			r = dref;
+			break;
+		case GIT_REF_OID:
+			r = ref;
+			break;
+		default:
+			continue;
+		}
+		if (!git_reference_target(r) ||
+		    git_reference_peel(&obj, r, GIT_OBJ_ANY))
+			goto err;
+		if (!(id = git_object_id(obj)))
+			goto err;
+		if (!(ci = commitinfo_getbyoid(id)))
+			break;
+
+		if (!(ris = reallocarray(ris, refcount + 1, sizeof(*ris))))
+			err(1, "realloc");
+		ris[refcount].ci = ci;
+		ris[refcount].ref = r;
+		refcount++;
+
+		git_object_free(obj);
+		obj = NULL;
+		git_reference_free(dref);
+		dref = NULL;
+	}
+	git_reference_iterator_free(it);
+
+	/* sort by type, date then shorthand name */
+	qsort(ris, refcount, sizeof(*ris), refs_cmp);
+
+	*pris = ris;
+	*prefcount = refcount;
+
+	return 0;
+
+err:
+	git_object_free(obj);
+	git_reference_free(dref);
+	commitinfo_free(ci);
+	for (i = 0; i < refcount; i++) {
+		commitinfo_free(ris[i].ci);
+		git_reference_free(ris[i].ref);
+	}
+	free(ris);
+
+	return -1;
+}
+
 FILE *
-efopen(const char *name, const char *flags)
+efopen(const char *filename, const char *flags)
 {
 	FILE *fp;
 
-	if (!(fp = fopen(name, flags)))
-		err(1, "fopen: '%s'", name);
+	if (!(fp = fopen(filename, flags)))
+		err(1, "fopen: '%s'", filename);
 
 	return fp;
 }
@@ -312,7 +441,7 @@ xmlencode(FILE *fp, const char *s, size_t len)
 		case '\'': fputs("&#39;",  fp); break;
 		case '&':  fputs("&amp;",  fp); break;
 		case '"':  fputs("&quot;", fp); break;
-		default:   fputc(*s, fp);
+		default:   putc(*s, fp);
 		}
 	}
 }
@@ -326,12 +455,12 @@ gphtextnl(FILE *fp, const char *s, size_t len)
 	for (i = 0; s[i] && i < len; i++) {
 		/* escape with 't' at the start of a line */
 		if (!n && (s[i] == 't' || s[i] == '['))
-			fputc('t', fp);
+			putc('t', fp);
 
 		switch (s[i]) {
 		case '\t': fputs("        ", fp);
 		case '\r': break;
-		default: fputc(s[i], fp);
+		default: putc(s[i], fp);
 		}
 		n = (s[i] != '\n');
 	}
@@ -353,7 +482,7 @@ gphtext(FILE *fp, const char *s, size_t len)
 			fputs("        ", fp);
 			break;
 		default:
-			fputc(*s, fp);
+			putc(*s, fp);
 			break;
 		}
 	}
@@ -377,7 +506,7 @@ gphlink(FILE *fp, const char *s, size_t len)
 			fputs("\\|", fp);
 			break;
 		default:
-			fputc(*s, fp);
+			putc(*s, fp);
 			break;
 		}
 	}
@@ -453,7 +582,7 @@ printtimeshort(FILE *fp, const git_time *intime)
 void
 writeheader(FILE *fp, const char *title)
 {
-	fputc('t', fp);
+	putc('t', fp);
 	gphtext(fp, title, strlen(title));
 	if (title[0] && strippedname[0])
 		fputs(" - ", fp);
@@ -489,16 +618,16 @@ writefooter(FILE *fp)
 {
 }
 
-int
+size_t
 writeblobgph(FILE *fp, const git_blob *blob)
 {
-	size_t n = 0, i, j, prev;
-	const char *nfmt = "%6d ";
+	size_t n = 0, i, j, len, prev;
+	const char *nfmt = "%6zu ";
 	const char *s = git_blob_rawcontent(blob);
-	git_off_t len = git_blob_rawsize(blob);
 
+	len = git_blob_rawsize(blob);
 	if (len > 0) {
-		for (i = 0, prev = 0; i < (size_t)len; i++) {
+		for (i = 0, prev = 0; i < len; i++) {
 			if (s[i] != '\n')
 				continue;
 			n++;
@@ -507,7 +636,7 @@ writeblobgph(FILE *fp, const git_blob *blob)
 				switch (s[j]) {
 				case '\r': break;
 				case '\t': fputs("        ", fp); break;
-				default: fputc(s[j], fp);
+				default: putc(s[j], fp);
 				}
 			}
 			prev = i + 1;
@@ -520,7 +649,7 @@ writeblobgph(FILE *fp, const git_blob *blob)
 				switch (s[j]) {
 				case '\r': break;
 				case '\t': fputs("        ", fp); break;
-				default: fputc(s[j], fp);
+				default: putc(s[j], fp);
 				}
 			}
 		}
@@ -549,12 +678,12 @@ printcommit(FILE *fp, struct commitinfo *ci)
 		fputs("|server|port]\n", fp);
 		fputs("Date:   ", fp);
 		printtime(fp, &(ci->author->when));
-		fputc('\n', fp);
+		putc('\n', fp);
 	}
 	if (ci->msg) {
-		fputc('\n', fp);
+		putc('\n', fp);
 		gphtextnl(fp, ci->msg, strlen(ci->msg));
-		fputc('\n', fp);
+		putc('\n', fp);
 	}
 }
 
@@ -656,9 +785,9 @@ printshowfile(FILE *fp, struct commitinfo *ci)
 			if (git_patch_get_hunk(&hunk, &nhunklines, patch, j))
 				break;
 
-			fputc('t', fp);
+			putc('t', fp);
 			gphtext(fp, hunk->header, hunk->header_len);
-			fputc('\n', fp);
+			putc('\n', fp);
 
 			for (k = 0; ; k++) {
 				if (git_patch_get_line_in_hunk(&line, patch, j, k))
@@ -670,7 +799,7 @@ printshowfile(FILE *fp, struct commitinfo *ci)
 				else
 					fputs(" ", fp);
 				gphtext(fp, line->content, line->content_len);
-				fputc('\n', fp);
+				putc('\n', fp);
 			}
 		}
 	}
@@ -760,7 +889,7 @@ err:
 }
 
 void
-printcommitatom(FILE *fp, struct commitinfo *ci)
+printcommitatom(FILE *fp, struct commitinfo *ci, const char *tag)
 {
 	fputs("<entry>\n", fp);
 
@@ -777,11 +906,16 @@ printcommitatom(FILE *fp, struct commitinfo *ci)
 	}
 	if (ci->summary) {
 		fputs("<title type=\"text\">", fp);
+		if (tag && tag[0]) {
+			fputs("[", fp);
+			xmlencode(fp, tag, strlen(tag));
+			fputs("] ", fp);
+		}
 		xmlencode(fp, ci->summary, strlen(ci->summary));
 		fputs("</title>\n", fp);
 	}
-	fprintf(fp, "<link rel=\"alternate\" type=\"text/html\" href=\"commit/%s.gph\" />\n",
-	        ci->oid);
+	fprintf(fp, "<link rel=\"alternate\" href=\"%scommit/%s.gph\" />\n",
+	        baseurl, ci->oid);
 
 	if (ci->author) {
 		fputs("<author>\n<name>", fp);
@@ -802,18 +936,20 @@ printcommitatom(FILE *fp, struct commitinfo *ci)
 		xmlencode(fp, ci->author->email, strlen(ci->author->email));
 		fputs("&gt;\nDate:   ", fp);
 		printtime(fp, &(ci->author->when));
-		fputc('\n', fp);
+		putc('\n', fp);
 	}
 	if (ci->msg) {
-		fputc('\n', fp);
+		putc('\n', fp);
 		xmlencode(fp, ci->msg, strlen(ci->msg));
 	}
 	fputs("\n</content>\n</entry>\n", fp);
 }
 
 int
-writeatom(FILE *fp)
+writeatom(FILE *fp, int all)
 {
+	struct referenceinfo *ris = NULL;
+	size_t refcount = 0;
 	struct commitinfo *ci;
 	git_revwalk *w = NULL;
 	git_oid id;
@@ -826,28 +962,41 @@ writeatom(FILE *fp)
 	xmlencode(fp, description, strlen(description));
 	fputs("</subtitle>\n", fp);
 
-	git_revwalk_new(&w, repo);
-	git_revwalk_push_head(w);
-	git_revwalk_simplify_first_parent(w);
+	/* all commits or only tags? */
+	if (all) {
+		git_revwalk_new(&w, repo);
+		git_revwalk_push_head(w);
+		git_revwalk_simplify_first_parent(w);
+		for (i = 0; i < m && !git_revwalk_next(&id, w); i++) {
+			if (!(ci = commitinfo_getbyoid(&id)))
+				break;
+			printcommitatom(fp, ci, "");
+			commitinfo_free(ci);
+		}
+		git_revwalk_free(w);
+	} else if (getrefs(&ris, &refcount) != -1) {
+		/* references: tags */
+		for (i = 0; i < refcount; i++) {
+			if (git_reference_is_tag(ris[i].ref))
+				printcommitatom(fp, ris[i].ci,
+				                git_reference_shorthand(ris[i].ref));
 
-	for (i = 0; i < m && !git_revwalk_next(&id, w); i++) {
-		if (!(ci = commitinfo_getbyoid(&id)))
-			break;
-		printcommitatom(fp, ci);
-		commitinfo_free(ci);
+			commitinfo_free(ris[i].ci);
+			git_reference_free(ris[i].ref);
+		}
+		free(ris);
 	}
-	git_revwalk_free(w);
 
 	fputs("</feed>\n", fp);
 
 	return 0;
 }
 
-int
-writeblob(git_object *obj, const char *fpath, const char *filename, git_off_t filesize)
+size_t
+writeblob(git_object *obj, const char *fpath, const char *filename, size_t filesize)
 {
 	char tmp[PATH_MAX] = "", *d;
-	int lc = 0;
+	size_t lc = 0;
 	FILE *fp;
 
 	if (strlcpy(tmp, fpath, sizeof(tmp)) >= sizeof(tmp))
@@ -859,9 +1008,9 @@ writeblob(git_object *obj, const char *fpath, const char *filename, git_off_t fi
 
 	fp = efopen(fpath, "w");
 	writeheader(fp, filename);
-	fputc('t', fp);
+	putc('t', fp);
 	gphtext(fp, filename, strlen(filename));
-	fprintf(fp, " (%juB)\n", (uintmax_t)filesize);
+	fprintf(fp, " (%zuB)\n", filesize);
 	fputs("---\n", fp);
 
 	if (git_blob_is_binary((git_blob *)obj)) {
@@ -923,13 +1072,11 @@ int
 writefilestree(FILE *fp, git_tree *tree, const char *path)
 {
 	const git_tree_entry *entry = NULL;
-	git_submodule *module = NULL;
 	git_object *obj = NULL;
-	git_off_t filesize;
 	const char *entryname;
-	char buf[256], filepath[PATH_MAX], entrypath[PATH_MAX];
-	size_t count, i;
-	int lc, r, ret;
+	char buf[256], filepath[PATH_MAX], entrypath[PATH_MAX], oid[8];
+	size_t count, i, lc, filesize;
+	int r, ret;
 
 	count = git_tree_entrycount(tree);
 	for (i = 0; i < count; i++) {
@@ -970,20 +1117,22 @@ writefilestree(FILE *fp, git_tree *tree, const char *path)
 			gphlink(fp, buf, strlen(buf));
 			fputs("  ", fp);
 			if (lc > 0)
-				fprintf(fp, "%7dL", lc);
+				fprintf(fp, "%7zuL", lc);
 			else
-				fprintf(fp, "%7juB", (uintmax_t)filesize);
+				fprintf(fp, "%7zuB", filesize);
 			fprintf(fp, "|%s/", relpath);
 			gphlink(fp, filepath, strlen(filepath));
 			fputs("|server|port]\n", fp);
 			git_object_free(obj);
-		} else if (!git_submodule_lookup(&module, repo, entryname)) {
+		} else if (git_tree_entry_type(entry) == GIT_OBJ_COMMIT) {
+			/* commit object in tree is a submodule */
 			fputs("[1|m---------  ", fp);
-			utf8pad(buf, sizeof(buf), entrypath, 50, ' ');
-			gphlink(fp, buf, strlen(buf));
+			gphlink(fp, entrypath, strlen(entrypath));
+			fputs(" @ ", fp);
+			git_oid_tostr(oid, sizeof(oid), git_tree_entry_id(entry));
+			gphlink(fp, oid, strlen(oid));
 			fprintf(fp, "|%s/file/.gitmodules.gph|server|port]\n", relpath);
 			/* NOTE: linecount omitted */
-			git_submodule_free(module);
 		}
 	}
 
@@ -1011,116 +1160,63 @@ writefiles(FILE *fp, const git_oid *id)
 	return ret;
 }
 
-int
-refs_cmp(const void *v1, const void *v2)
-{
-	git_reference *r1 = (*(git_reference **)v1);
-	git_reference *r2 = (*(git_reference **)v2);
-	int r;
-
-	if ((r = git_reference_is_branch(r1) - git_reference_is_branch(r2)))
-		return r;
-
-	return strcmp(git_reference_shorthand(r1),
-	              git_reference_shorthand(r2));
-}
-
 int
 writerefs(FILE *fp)
 {
+	struct referenceinfo *ris = NULL;
 	struct commitinfo *ci;
-	const git_oid *id = NULL;
-	git_object *obj = NULL;
-	git_reference *dref = NULL, *r, *ref = NULL;
-	git_reference_iterator *it = NULL;
-	git_reference **refs = NULL;
 	size_t count, i, j, refcount;
 	const char *titles[] = { "Branches", "Tags" };
-	const char *name;
+	const char *s;
 	char buf[256];
 
-	if (git_reference_iterator_new(&it, repo))
+	if (getrefs(&ris, &refcount) == -1)
 		return -1;
 
-	for (refcount = 0; !git_reference_next(&ref, it); refcount++) {
-		if (!(refs = reallocarray(refs, refcount + 1, sizeof(git_reference *))))
-			err(1, "realloc");
-		refs[refcount] = ref;
-	}
-	git_reference_iterator_free(it);
-
-	/* sort by type then shorthand name */
-	qsort(refs, refcount, sizeof(git_reference *), refs_cmp);
-
-	for (j = 0; j < 2; j++) {
-		for (i = 0, count = 0; i < refcount; i++) {
-			if (!(git_reference_is_branch(refs[i]) && j == 0) &&
-			    !(git_reference_is_tag(refs[i]) && j == 1))
-				continue;
-
-			switch (git_reference_type(refs[i])) {
-			case GIT_REF_SYMBOLIC:
-				if (git_reference_resolve(&dref, refs[i]))
-					goto err;
-				r = dref;
-				break;
-			case GIT_REF_OID:
-				r = refs[i];
-				break;
-			default:
-				continue;
-			}
-			if (!git_reference_target(r) ||
-			    git_reference_peel(&obj, r, GIT_OBJ_ANY))
-				goto err;
-			if (!(id = git_object_id(obj)))
-				goto err;
-			if (!(ci = commitinfo_getbyoid(id)))
-				break;
-
-			/* print header if it has an entry (first). */
-			if (++count == 1) {
-				fprintf(fp, "%s\n", titles[j]);
-				fprintf(fp, "  %-32.32s", "Name");
-				fprintf(fp, "  %-16.16s", "Last commit date");
-				fprintf(fp, "  %s\n", "Author");
-			}
-
-			name = git_reference_shorthand(r);
-
-			fputs("  ", fp);
-			utf8pad(buf, sizeof(buf), name, 32, ' ');
-			gphlink(fp, buf, strlen(buf));
-			fputs("  ", fp);
-			if (ci->author)
-				printtimeshort(fp, &(ci->author->when));
-			else
-				fputs("                ", fp);
-			fputs("  ", fp);
-			if (ci->author) {
-				utf8pad(buf, sizeof(buf), ci->author->name, 25, '\0');
-				gphlink(fp, buf, strlen(buf));
-			}
-			fputs("\n", fp);
-
-			commitinfo_free(ci);
-			git_object_free(obj);
-			obj = NULL;
-			git_reference_free(dref);
-			dref = NULL;
+	for (i = 0, j = 0, count = 0; i < refcount; i++) {
+		if (j == 0 && git_reference_is_tag(ris[i].ref)) {
+			/* table footer */
+			if (count)
+				fputs("\n", fp);
+			count = 0;
+			j = 1;
 		}
-		/* table footer */
-		if (count)
-			fputs("\n", fp);
+
+		/* print header if it has an entry (first). */
+		if (++count == 1) {
+			fprintf(fp, "%s\n", titles[j]);
+			fprintf(fp, "  %-32.32s", "Name");
+			fprintf(fp, "  %-16.16s", "Last commit date");
+			fprintf(fp, "  %s\n", "Author");
+		}
+
+		ci = ris[i].ci;
+		s = git_reference_shorthand(ris[i].ref);
+
+		fputs("  ", fp);
+		utf8pad(buf, sizeof(buf), s, 32, ' ');
+		gphlink(fp, buf, strlen(buf));
+		fputs("  ", fp);
+		if (ci->author)
+			printtimeshort(fp, &(ci->author->when));
+		else
+			fputs("                ", fp);
+		fputs("  ", fp);
+		if (ci->author) {
+			utf8pad(buf, sizeof(buf), ci->author->name, 25, '\0');
+			gphlink(fp, buf, strlen(buf));
+		}
+		fputs("\n", fp);
 	}
+	/* table footer */
+	if (count)
+		fputs("\n", fp);
 
-err:
-	git_object_free(obj);
-	git_reference_free(dref);
-
-	for (i = 0; i < refcount; i++)
-		git_reference_free(refs[i]);
-	free(refs);
+	for (i = 0; i < refcount; i++) {
+		commitinfo_free(ris[i].ci);
+		git_reference_free(ris[i].ref);
+	}
+	free(ris);
 
 	return 0;
 }
@@ -1128,7 +1224,8 @@ err:
 void
 usage(char *argv0)
 {
-	fprintf(stderr, "%s [-b baseprefix] [-c cachefile | -l commits] repodir\n", argv0);
+	fprintf(stderr, "%s [-b baseprefix] [-c cachefile | -l commits] "
+	        "[-u baseurl] repodir\n", argv0);
 	exit(1);
 }
 
@@ -1167,6 +1264,10 @@ main(int argc, char *argv[])
 			if (argv[i][0] == '\0' || *p != '\0' ||
 			    nlogcommits <= 0 || errno)
 				usage(argv[0]);
+		} else if (argv[i][1] == 'u') {
+			if (i + 1 >= argc)
+				usage(argv[0]);
+			baseurl = argv[++i];
 		}
 	}
 	if (!repodir)
@@ -1244,7 +1345,7 @@ main(int argc, char *argv[])
 	}
 
 	/* check LICENSE */
-	for (i = 0; i < sizeof(licensefiles) / sizeof(*licensefiles) && !license; i++) {
+	for (i = 0; i < LEN(licensefiles) && !license; i++) {
 		if (!git_revparse_single(&obj, repo, licensefiles[i]) &&
 		    git_object_type(obj) == GIT_OBJ_BLOB)
 			license = licensefiles[i] + strlen("HEAD:");
@@ -1252,7 +1353,7 @@ main(int argc, char *argv[])
 	}
 
 	/* check README */
-	for (i = 0; i < sizeof(readmefiles) / sizeof(*readmefiles) && !readme; i++) {
+	for (i = 0; i < LEN(readmefiles) && !readme; i++) {
 		if (!git_revparse_single(&obj, repo, readmefiles[i]) &&
 		    git_object_type(obj) == GIT_OBJ_BLOB)
 			readme = readmefiles[i] + strlen("HEAD:");
@@ -1310,7 +1411,9 @@ main(int argc, char *argv[])
 		if (head)
 			writelog(fp, head);
 	}
-	fprintf(fp, "\n[0|Atom feed|%s/atom.xml|server|port]\n", relpath);
+	fputs("\n", fp);
+	fprintf(fp, "[0|Atom feed|%s/atom.xml|server|port]\n", relpath);
+	fprintf(fp, "[0|Atom feed (tags)|%s/tags.xml|server|port]\n", relpath);
 	writefooter(fp);
 	fclose(fp);
 
@@ -1331,7 +1434,12 @@ main(int argc, char *argv[])
 
 	/* Atom feed */
 	fp = efopen("atom.xml", "w");
-	writeatom(fp);
+	writeatom(fp, 1);
+	fclose(fp);
+
+	/* Atom feed for tags / releases */
+	fp = efopen("tags.xml", "w");
+	writeatom(fp, 0);
 	fclose(fp);
 
 	/* rename new cache file on success */
